from __future__ import absolute_import, unicode_literals

import codecs
import os
import re
import string
import sys
import unicodedata

import cucco.regex as regex

from cucco.config import Config

PATH = os.path.dirname(__file__)


class Cucco(object):
    """This class offers methods for text normalization.

    Attributes:
        config: Config to use.
        lazy_load: Whether or not to lazy load files.
    """
    __punctuation = set(string.punctuation)

    def __init__(
            self,
            config=None,
            lazy_load=False):
        self._config = config if config else Config()

        self._characters_regexes = dict()
        self._logger = self._config.logger
        self.__stop_words = dict()

        # Load stop words
        self._load_stop_words(self._config.language if lazy_load else None)

    def _load_stop_words(self, language=None):
        """Load stop words into __stop_words set.

        Stop words will be loaded according to the language code
        received during instantiation.

        Args:
            language: Language code.

        Returns:
            A boolean indicating if the file was loaded.
        """
        self._logger.debug('Loading stop words')

        loaded = False

        if language:
            file_path = 'data/stop-' + language
            loaded = self._parse_stop_words_file(os.path.join(PATH, file_path))
        else:
            for file in os.listdir(os.path.join(PATH, 'data')):
                loaded = self._parse_stop_words_file(os.path.join(PATH, 'data', file)) or loaded

        return loaded

    @staticmethod
    def _parse_normalizations(normalizations):
        """Parse and yield normalizations.

        Parse normalizations parameter yielding all normalizations and
        arguments found on it.

        Args:
            normalizations: List of normalizations.

        Yields:
            A tuple with a parsed normalization. The first item will
            contain the normalization name and the second will be a dict
            with the arguments to use for the normalization.
        """
        str_type = str if sys.version_info[0] > 2 else (str, unicode)

        for normalization in normalizations:
            yield (normalization, {}) if isinstance(normalization, str_type) else normalization

    def _parse_stop_words_file(self, path):
        """Load stop words from the given path.

        Parse an stop words file, saving each word found in it in a set
        for the languaage of the file. This language is obtained from
        the file name. If the file doesn't exist, the method will have
        no effect.

        Args:
            path: Path to the stop words file.

        Returns:
            A boolean indicating if the file was loaded.
        """
        language = None
        loaded = False

        if os.path.isfile(path):
            self._logger.debug('Loading stop words in %s', path)

            language = path.split('-')[-1]

            if not language in self.__stop_words:
                self.__stop_words[language] = set()

            with codecs.open(path, 'r', 'UTF-8') as file:
                loaded = True
                for word in file:
                    self.__stop_words[language].add(word.strip())

        return loaded

    def normalize(self, text, normalizations=None):
        """Normalize a given text applying all normalizations.

        Normalizations to apply can be specified through a list
        parameter and will be executed in the same order.

        Args:
            text: The text to be processed.
            normalizations: List of normalizations to apply.

        Returns:
            The text normalized.
        """
        for normalization, kwargs in self._parse_normalizations(
                normalizations or self._config.normalizations):
            text = getattr(self, normalization)(text, **kwargs)

        return text

    @staticmethod
    def remove_accent_marks(text, excluded=None):
        """Remove accent marks from input text.

        This function remove accent marks in the text, leaving
        untouched unicode characters defined in the 'excluded'
        parameter.

        Args:
            text: The text to be processed.
            excluded: Set of unicode characters to exclude.

        Returns:
            The text without accent marks.
        """
        if excluded is None:
            excluded = set()

        return unicodedata.normalize(
            'NFKC', ''.join(
                c for c in unicodedata.normalize(
                    'NFKD', text) if unicodedata.category(c) != 'Mn' or c in excluded))

    @staticmethod
    def remove_extra_whitespaces(text):
        """Remove extra white spaces from input text.

        This function removes white spaces from the beginning and
        the end of the string, but also duplicated white spaces
        between words.

        Args:
            text: The text to be processed.

        Returns:
            The text without extra white spaces.
        """
        return ' '.join(text.split())

    def remove_stop_words(self, text, ignore_case=True, language=None):
        """Remove stop words.

        Stop words are loaded on class instantiation according
        with the specified language.

        Args:
            text: The text to be processed.
            ignore_case: Whether or not ignore case.
            language: Code of the language to use.

        Returns:
            The text without stop words.
        """
        if not language:
            language = self._config.language

        if language not in self.__stop_words:
            if not self._load_stop_words(language):
                self._logger.error('No stop words file for the given language')
                return text

        return ' '.join(word for word in text.split(' ') if (
            word.lower() if ignore_case else word) not in self.__stop_words[language])

    def replace_characters(self, text, characters, replacement=''):
        """Remove characters from text.

        Remove custom characters from input text or replace them
        with a string if specified.

        Args:
            text: The text to be processed.
            characters: Characters that will be replaced.
            replacement: New text that will replace the custom characters.

        Returns:
            The text without the given characters.
        """
        if not characters:
            return text

        characters = ''.join(sorted(characters))
        if characters in self._characters_regexes:
            characters_regex = self._characters_regexes[characters]
        else:
            characters_regex = re.compile("[%s]" % re.escape(characters))
            self._characters_regexes[characters] = characters_regex

        return characters_regex.sub(replacement, text)

    @staticmethod
    def replace_emails(text, replacement=''):
        """Remove emails address from text.

        Remove email addresses from input text or replace them
        with a string if specified.

        Args:
            text: The text to be processed.
            replacement: New text that will replace email addresses.

        Returns:
            The text without email addresses.
        """
        return re.sub(regex.EMAIL_REGEX, replacement, text)

    @staticmethod
    def replace_emojis(text, replacement=''):
        """Remove emojis from text.

        Remove emojis from input text or replace them with a
        string if specified.

        Args:
            text: The text to be processed.
            replacement: New text that will replace emojis.

        Returns:
            The text without hyphens.
        """
        return regex.EMOJI_REGEX.sub(replacement, text)

    @staticmethod
    def replace_hyphens(text, replacement=' '):
        """Replace hyphens in text.

        Replace hyphens from input text with a whitespace or a
        string if specified.

        Args:
            text: The text to be processed.
            replacement: New text that will replace hyphens.

        Returns:
            The text without hyphens.
        """
        return text.replace('-', replacement)

    def replace_punctuation(self, text, excluded=None, replacement=''):
        """Replace punctuation symbols in text.

        Remove punctuation from input text or replace them with a
        string if specified. Characters replaces will be those
        in string.punctuation.

        Args:
            text: The text to be processed.
            excluded: Set of characters to exclude.
            replacement: New text that will replace punctuation.

        Returns:
            The text without punctuation.
        """
        if excluded is None:
            excluded = set()
        elif not isinstance(excluded, set):
            excluded = set(excluded)
        punct = ''.join(self.__punctuation.difference(excluded))

        return self.replace_characters(
            text, characters=punct, replacement=replacement)

    @staticmethod
    def replace_symbols(
            text,
            form='NFKD',
            excluded=None,
            replacement=''):
        """Replace symbols in text.

        Remove symbols from input text or replace them with a
        string if specified.

        Args:
            text: The text to be processed.
            form: Unicode form.
            excluded: Set of unicode characters to exclude.
            replacement: New text that will replace symbols.

        Returns:
            The text without symbols.
        """
        if excluded is None:
            excluded = set()

        categories = set(['Mn', 'Sc', 'Sk', 'Sm', 'So'])

        return ''.join(c if unicodedata.category(c) not in categories or c in excluded
                       else replacement for c in unicodedata.normalize(form, text))

    @staticmethod
    def replace_urls(text, replacement=''):
        """Replace URLs in text.

        Remove URLs from input text or replace them with a
        string if specified.

        Args:
            text: The text to be processed.
            replacement: New text that will replace URLs.

        Returns:
            The text without URLs.
        """
        return re.sub(regex.URL_REGEX, replacement, text)



from __future__ import absolute_import

import os
import time

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

BATCH_EXTENSION = '.cucco'

def files_generator(path, recursive):
    """Yield files found in a given path.

    Walk over a given path finding and yielding all
    files found on it. This can be done only on the root
    directory or recursively.

    Args:
        path: Path to the directory.
        recursive: Whether to find files recursively or not.

    Yields:
        A tuple for each file in the given path containing
        the path and the name of the file.
    """
    if recursive:
        for (path, _, files) in os.walk(path):
            for file in files:
                if not file.endswith(BATCH_EXTENSION):
                    yield (path, file)
    else:
        for file in os.listdir(path):
            if (os.path.isfile(os.path.join(path, file)) and
                    not file.endswith(BATCH_EXTENSION)):
                yield (path, file)

def lines_generator(path):
    """Yield lines in a given file.

    Iterates over a file lines yielding them.

    Args:
        path: Path to the file.

    Yields:
        Lines on a given file.
    """
    with open(path, 'r') as file:
        for line in file:
            yield line


class Batch(object):
    """Class to apply normalizations in batch mode.

    This class permits to apply normalizations over a group
    of files. It counts with two modes. The first one works
    over all the files in a directory and the second one
    watch for new files in a given path. Both modes generate
    new files with the result of the normalizations, letting
    the original files unchanged.

    Attributes:
        config: Config to use.
        cucco: Reference to cucco object.
   """

    def	__init__(self, config, cucco):
        """Inits Batch class."""
        self._config = config
        self._cucco = cucco
        self._logger = config.logger
        self._observer = None
        self._watch = False

    def _process_file(self, path):
        """Process a file applying normalizations.

        Get a file as input and generate a new file with the
        result of applying normalizations to every single line
        in the original file. The extension for the new file
        will be the one defined in BATCH_EXTENSION.

        Args:
            path: Path to the file.
        """
        if self._config.verbose:
            self._logger.info('Processing file "%s"', path)

        output_path = '%s%s' % (path, BATCH_EXTENSION)

        with open(output_path, 'w') as file:
            for line in lines_generator(path):
                file.write('%s\n' % self._cucco.normalize(line))

        if self._config.debug:
            self._logger.debug('Created file "%s"', output_path)

    def process_files(self, path, recursive=False):
        """Apply normalizations over all files in the given directory.

        Iterate over all files in a given directory. Normalizations
        will be applied to each file, storing the result in a new file.
        The extension for the new file will be the one defined in
        BATCH_EXTENSION.

        Args:
            path: Path to the directory.
            recursive: Whether to find files recursively or not.
        """
        self._logger.info('Processing files in "%s"', path)

        for (path, file) in files_generator(path, recursive):
            if not file.endswith(BATCH_EXTENSION):
                self._process_file(os.path.join(path, file))

    def stop_watching(self):
        """Stop watching for files.

        Stop the observer started by watch function and finish
        thread life.
        """
        self._watch = False

        if self._observer:
            self._logger.info('Stopping watcher')
            self._observer.stop()
            self._logger.info('Watcher stopped')

    def watch(self, path, recursive=False):
        """Watch for files in a directory and apply normalizations.

        Watch for new or changed files in a directory and apply
        normalizations over them.

        Args:
            path: Path to the directory.
            recursive: Whether to find files recursively or not.
        """
        self._logger.info('Initializing watcher for path "%s"', path)

        handler = FileHandler(self)
        self._observer = Observer()
        self._observer.schedule(handler, path, recursive)

        self._logger.info('Starting watcher')
        self._observer.start()
        self._watch = True

        try:
            self._logger.info('Waiting for file events')
            while self._watch:
                time.sleep(1)
        except KeyboardInterrupt: # pragma: no cover
            self.stop_watching()

        self._observer.join()


class FileHandler(FileSystemEventHandler):
    """Handler to use by Batch watcher.

    This class is used by Batch's watch mode. The handler will
    listen for new and changed files.

    Attributes:
        batch: Reference to Batch object.
    """

    def __init__(self, batch):
        """Inits Batch class."""
        self._batch = batch
        self._debug = batch._config.debug
        self._logger = batch._logger

    def _process_event(self, event):
        """Process received events.

        Process events received, applying normalization for those
        events referencing a new or changed file and only if it's
        not the result of a previous normalization.

        Args:
            event: Event to process.
        """
        if (not event.is_directory and
                not event.src_path.endswith(BATCH_EXTENSION)):
            self._logger.info('Detected file change: %s', event.src_path)
            self._batch._process_file(event.src_path)

    def on_created(self, event):
        """Function called everytime a new file is created.

        Args:
            event: Event to process.
        """
        if self._debug:
            self._logger.debug('Detected create event on watched path: %s', event.src_path)

        self._process_event(event)

    def on_modified(self, event):
        """Function called everytime a new file is modified.

        Args:
            event: Event to process.
        """
        if self._debug:
            self._logger.debug('Detected modify event on watched path: %s', event.src_path)

        self._process_event(event)



#-*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import re

"""
Regular expression to match URLs as seen on http://daringfireball.net/2010/07/improved_regex_for_matching_urls
"""
URL_REGEX = re.compile(
    r'(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’]))',
    re.IGNORECASE)

"""
Regular expression to match email addresses as seen on http://www.wellho.net/resources/ex.php4?item=y115/relib.py
"""
EMAIL_REGEX = re.compile(r"[-a-z0-9_.]+@(?:[-a-z0-9]+\.)+[a-z]{2,6}", re.IGNORECASE)

try:
    EMOJI_REGEX = re.compile(u'([\U00002600-\U000027BF])|([\U0001f300-\U0001f64F])|([\U0001f680-\U0001f6FF])')
except re.error: # pragma: no cover
    EMOJI_REGEX = re.compile(u'([\u2600-\u27BF])|([\uD83C][\uDF00-\uDFFF])|([\uD83D][\uDC00-\uDE4F])|([\uD83D][\uDE80-\uDEFF])')



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import click
import sys

import cucco.logging as logging

from cucco.batch import Batch
from cucco.config import Config
from cucco.cucco import Cucco
from cucco.errors import ConfigError

@click.command()
@click.argument('path')
@click.option('--recursive', '-r', is_flag=True,
              help='Whether to search for files recursively.')
@click.option('--watch', '-w', is_flag=True,
              help='Watch for new files in the given path.')
@click.pass_context
def batch(ctx, path, recursive, watch):
    """
    Normalize files in a path.

    Apply normalizations over all files found in a given path.
    The normalizations applied will be those defined in the config
    file. If no config is specified, the default normalizations will
    be used.
    """
    batch = Batch(ctx.obj['config'], ctx.obj['cucco'])

    if watch:
        batch.watch(path, recursive)
    else:
        batch.process_files(path, recursive)

@click.command()
@click.argument('text', required=False)
@click.pass_context
def normalize(ctx, text):
    """
    Normalize text or piped input.

    Normalize text passed as an argument to this command using
    the specified config (default values if --config option is
    not used).

    Pipes can be used along this command to process the output
    of another cli. This is the default behaviour when no text
    is defined.
    """
    if text:
        click.echo(ctx.obj['cucco'].normalize(text))
    else:
        for line in sys.stdin:
            click.echo(ctx.obj['cucco'].normalize(line))

@click.group()
@click.option('--config', '-c',
              help='Path to config file.')
@click.option('--debug', '-d', is_flag=True,
              help='Show debug messages.')
@click.option('--language', '-l', default='en',
              help='Language to use for the normalization.')
@click.option('--verbose', '-v', is_flag=True,
              help='Increase output verbosity.')
@click.version_option()
@click.pass_context
def cli(ctx, config, debug, language, verbose):
    """
    Cucco allows to apply normalizations to a given text or file.
    This normalizations include, among others, removal of accent
    marks, stop words an extra white spaces, replacement of
    punctuation symbols, emails, emojis, etc.

    For more info on how to use and configure Cucco, check the
    project website at https://cucco.io.
    """
    ctx.obj = {}

    try:
        ctx.obj['config'] = Config(normalizations=config,
                                   language=language,
                                   debug=debug,
                                   verbose=verbose)
    except ConfigError as e:
        click.echo(e.message)
        sys.exit(-1)

    ctx.obj['cucco'] = Cucco(ctx.obj['config'])

cli.add_command(batch)
cli.add_command(normalize)



#!/user/bin/env python

import re

from setuptools import setup, find_packages

version = re.search(
    '^__version__\s*=\s*\'(.*)\'',
    open('cucco/__init__.py').read(),
    re.M).group(1)

setup(name='cucco',
      version=version,
      description='Python library for text normalization',
      author='David Moreno-Garcia',
      author_email='david.mogar@gmail.com',
      license='MIT',
      url='https://github.com/davidmogar/cucco',
      download_url='https://github.com/davidmogar/cucco/tarball/' + version,
      keywords=['normalization', 'language', 'text', 'manipulation'],
      packages=find_packages(exclude=['tests']),
      entry_points={
          'console_scripts': [
              'cucco=cucco.cli:cli'
          ]
      },
      include_package_data=True,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Topic :: Software Development :: Libraries',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
      ]
)



from __future__ import absolute_import

import logging

def initialize_logger(debug):
    """Set up logger to be used by the library.

    Args:
        debug: Wheter to use debug level or not.

    Returns:
        A logger ready to be used.
    """
    level = logging.DEBUG if debug else logging.INFO
    logger = logging.getLogger('cucco')
    logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s %(levelname).1s %(message)s')
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger



class Error(Exception):
    """Base class for exceptions in this module."""
    pass

class ConfigError(Error):
    """Exception raised for errors loading config file.

    Attributes:
        message: Explanation of the error.
    """

    def __init__(self, message):
        self.message = message



from __future__ import absolute_import

import sys
import yaml

import cucco.logging as logging

from cucco.errors import ConfigError

DEFAULT_NORMALIZATIONS = [
    'replace_punctuation',
    'remove_extra_whitespaces',
    'replace_symbols',
    'remove_stop_words'
]
STR_TYPE = str if sys.version_info[0] > 2 else (str, unicode)


class Config(object):
    """Class to manage cucco configuration.

    This class allows to handle all cucco configuration and is
    used by the different modules.

    Attributes:
        debug: Whether to show debug messages or not.
        language: Language to be used for the normalizations.
        normalizations: List or path to config file.
        verbose: Level of output verbosity.
    """

    normalizations = DEFAULT_NORMALIZATIONS

    def	__init__(self,
                 normalizations=None,
                 language='en',
                 logger=None,
                 debug=False,
                 verbose=False):
        """Inits Config class."""
        self.debug = debug
        self.language = language
        self.logger = logger or logging.initialize_logger(debug)
        self.verbose = verbose or debug

        if normalizations:
            if isinstance(normalizations, STR_TYPE):
                normalizations = self._load_from_file(normalizations)

            self.normalizations = self._parse_normalizations(normalizations)

    @staticmethod
    def _load_from_file(path):
        """Load a config file from the given path.

        Load all normalizations from the config file received as
        argument. It expects to find a YAML file with a list of
        normalizations and arguments under the key 'normalizations'.

        Args:
            path: Path to YAML file.
        """
        config = []

        try:
            with open(path, 'r') as config_file:
                config = yaml.load(config_file)['normalizations']
        except EnvironmentError as e:
            raise ConfigError('Problem while loading file: %s' %
                              e.args[1] if len(e.args) > 1 else e)
        except (TypeError, KeyError) as e:
            raise ConfigError('Config file has an unexpected structure: %s' % e)
        except yaml.YAMLError:
            raise ConfigError('Invalid YAML file syntax')

        return config

    @staticmethod
    def _parse_normalization(normalization):
        """Parse a normalization item.

        Transform dicts into a tuple containing the normalization
        options. If a string is found, the actual value is used.

        Args:
            normalization: Normalization to parse.

        Returns:
            Tuple or string containing the parsed normalization.
        """
        parsed_normalization = None

        if isinstance(normalization, dict):
            if len(normalization.keys()) == 1:
                items = list(normalization.items())[0]
                if len(items) == 2: # Two elements tuple
                    # Convert to string if no normalization options
                    if items[1] and isinstance(items[1], dict):
                        parsed_normalization = items
                    else:
                        parsed_normalization = items[0]
        elif isinstance(normalization, STR_TYPE):
            parsed_normalization = normalization

        return parsed_normalization

    def _parse_normalizations(self, normalizations):
        """Returns a list of parsed normalizations.

        Iterates over a list of normalizations, removing those
        not correctly defined. It also transform complex items
        to have a common format (list of tuples and strings).

        Args:
            normalizations: List of normalizations to parse.

        Returns:
            A list of normalizations after being parsed and curated.
        """
        parsed_normalizations = []

        if isinstance(normalizations, list):
            for item in normalizations:
                normalization = self._parse_normalization(item)
                if normalization:
                    parsed_normalizations.append(normalization)
        else:
            raise ConfigError('List expected. Found %s' % type(normalizations))

        return parsed_normalizations
